/*
 * Copyright (C) 2005-2006  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "avrdebug_par.h"

#ifndef USE_LIBC5
# include <sys/io.h> // für ioperm(); glibc
#else
# include "unistd.h" // für ioperm(); libc5
#endif

#include <unistd.h>
#include <iostream>

#define USLEEP_TIME   7
#define HIGHMASK      0x7F00

/**
 * Setzt /SS low
 */
inline void avrdebug_par::selectSlave()
{
  outb(1<<3, uLPTPort+2);
  usleep(100);
}

/**
 * Setzt /SS high
 */
inline void avrdebug_par::deselectSlave()
{
  outb(0<<3, uLPTPort+2);
}

void avrdebug_par::dataExchange(unsigned char  uchToSend,
                                unsigned char& uchRec)
{
  uchRec = 0;

  for (unsigned char uchCnt = 0; uchCnt < 8; uchCnt++)
  {
    uchRec <<= 1;

    // set data pin
    if ((uchToSend & 128) != 0)
      uchDataState |= 128;
    else
      uchDataState &= 127;
    uchToSend <<= 1;
    outb(uchDataState, uLPTPort);

    usleep(USLEEP_TIME);

    // set clk pin
    uchDataState |= 1;
    outb(uchDataState, uLPTPort);

    usleep(USLEEP_TIME);

    // scan input pin
    if ((inb(uLPTPort+1) & 128) == 0)
      uchRec |= 1;

    // clear clock pin
    uchDataState &= 254;
    outb(uchDataState, uLPTPort);

    usleep(USLEEP_TIME);
  }
  uchDataState |= 128;
  outb(uchDataState, uLPTPort);
  usleep(USLEEP_TIME);
}

void avrdebug_par::executeCmd(DebugCmd      cmd,
                              ResultBlock&  results)
{
  if (cmd.write())
  {
    int           nData = cmd.value();
    unsigned char uchRec;

    int nAdrStart = cmd.addr();
    int nAdrEnd   = nAdrStart + cmd.size() -1;

    for (int nAdr=nAdrStart; nAdr<=nAdrEnd; nAdr++)
    {
      dataExchange(nAdr & 0xFF, uchRec);
      if (uchRec != 0)
      {
        std::cerr << "0 erwartet, " << (int)uchRec << " gelesen.\n";
        setErr("read something unexpected");
      }
      else
      {
        if ((nAdr & HIGHMASK) != nAdr)
        {
          setErr("address out of range");
        }
        
        dataExchange(((nAdr & HIGHMASK) >> 8) | 128, uchRec);
        if (uchRec != 1)
        {
          std::cerr << "1 erwartet, " << (int)uchRec << " gelesen.\n";
          setErr("read something unexpected");
        }
        else
          dataExchange(nData, uchRec);
      }

      results.push_back(uchRec);
    }
  }
  else
  {
    int nAdrStart = cmd.addr();
    int nAdrEnd   = nAdrStart + cmd.size() -1;

    for (int nAdr=nAdrStart; nAdr<=nAdrEnd; nAdr++)
    {
      unsigned char uchRec;

      dataExchange(nAdr & 0xFF, uchRec);
      if (uchRec != 0)
      {
        std::cerr << "0 erwartet, " << (int)uchRec << " gelesen.\n";
        setErr("read something unexpected");
      }
      
      if ((nAdr & HIGHMASK) != nAdr)
      {
        setErr("address out of range");
      }
        
      dataExchange((nAdr & HIGHMASK) >> 8, uchRec);
      if (uchRec != 1)
      {
        std::cerr << "1 erwartet, " << (int)uchRec << " gelesen.\n";
        setErr("read something unexpected");
      }
      dataExchange(0, uchRec);

      results.push_back(uchRec);
    }
  }
}

int avrdebug_par::resetTarget()
{
  // reset device
  std::cerr << "Führe Reset aus...\n";

  uchDataState = 128;
  outb(uchDataState, uLPTPort);

  usleep(100000);

  uchDataState = 128 + 2; // set /Reset
  outb(uchDataState, uLPTPort);
  
  return(0);
}

int avrdebug_par::connect()
{
  if (testCom())
  {
    if (!testCom())
    {
      setErr("could not sync to target");
      return(-1);
    }
    return(0);
  }
  else
  {
    setErr("could not sync to target");
    return(-1);
  }
}

avrdebug_par::avrdebug_par(int  nAdr)
{
  setOK();

  uLPTPort = nAdr;

  /* Get access to the ports */
  if (ioperm(uLPTPort,3,1))
  {
    std::cerr << "ioperm\n";
    setErr("ioperm");
  }
  else
  {
    deselectSlave();

    uchDataState = 128 + 2; // set /Reset
    outb(uchDataState, uLPTPort);
  }
}

avrdebug_par::~avrdebug_par()
{
  deselectSlave();

  /* We don't need the ports anymore */
  if (ioperm(uLPTPort,3,0))
  {
    std::cerr << "Error: ioperm\n";
  }
}

bool avrdebug_par::testCom()
  // Man muss beim ersten Austausch eine 0 erhalten, beim zweiten eine eins.
  // Beim dritten Mal kommt irgendwas.
  // Diese Version versucht außerdem, eine außer Takt geratene SPI-Kommunikation
  // wieder zu synchronisieren.
  // Man tauscht zwei Byte aus und sollte 0x00 und 0x01 erhalten.
  // Wenn man eine eins erhält, sollten also vorher mindestens 15 nullen
  // eingelaufen sein.
  // 00000000 00000001 XXXXXXXX 00000000 00000001 XXXXXXXX 00000000 00000001 XXXXXXXX
{
  unsigned char uchRec;
  unsigned int  cntBits, nOK;
  unsigned char uchCntNullen;
  unsigned int  cntByte3;

  selectSlave();

  std::cerr << "\n";
  cntBits = 64;
  nOK = 0;
  uchCntNullen = 0;
  cntByte3 = 0;

  while (cntBits > 0 && nOK != 2)
  {
    // clear data pin
    uchDataState &= 127;
    outb(uchDataState, uLPTPort);

    usleep(USLEEP_TIME);

    // set clk pin
    uchDataState |= 1;
    outb(uchDataState, uLPTPort);

    usleep(USLEEP_TIME);

    // scan input pin
    if (cntByte3)
    {
      // Es wird das dritte Byte eingelesen, das ist egal.
      cntByte3--;
      uchCntNullen = 0;
      std::cerr << "X";
      if (cntByte3 == 0)
      {
        std::cerr << "\n";
      }
    }
    else
    {
      if ((inb(uLPTPort+1) & 128) == 0)
      {
        std::cerr << "1 (" << (int)uchCntNullen << ") ";

        if (uchCntNullen < 15)
        {
          std::cerr << " zu wenig \n";
          nOK = 0;
        }
        else
        {
          if ((uchCntNullen >= 15 && nOK == 0)
              || (uchCntNullen == 15 && nOK == 1))
          {
            // Es sind vorher genügend nullen eingelaufen.
            cntByte3 = 8;
            nOK++;
            if (uchCntNullen == 15)
              std::cerr << " passend\n";
            else
              std::cerr << " genug\n";
          }
          else
          {
            // Fehler
            uchCntNullen = 0;
            std::cerr << " Fehler \n";
            nOK = 0;
          }
        }
        uchCntNullen = 0;
      }
      else
      {
        std::cerr << "0";
        uchCntNullen++;
      }
    }

    // clear clock pin
    uchDataState &= 254;
    outb(uchDataState, uLPTPort);

    usleep(USLEEP_TIME);

    cntBits--;
  }
  uchDataState |= 128;
  outb(uchDataState, uLPTPort);
  std::cerr << "\n";

  dataExchange(0, uchRec);
  std::cerr << "0 geschrieben, bekommen: " << (int)uchRec << "\n";

  if (nOK == 2)
  {
    std::cerr << "Offensichtlich kann man sich unterhalten...\n";
    return(true);
  }
  else
  {
    std::cerr << "Gespräch nicht wie erwartet.\n";
    return(false);
  }
}
