// -*- mode: c; mode: fold -*-
//

/*
 * Copyright (C) 2005-2007  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "avrdebug_ser.h"

#include <iostream>

#define VERBOSE_DBG 0

void avrdebug_ser::executeCmd(DebugCmd      cmd,
                              ResultBlock&  results) 
{
#if (VERBOSE_DBG != 0)
  std::cout << cmd.describe() << "\n";
#endif
  
  if (cmd.write())
  {
    unsigned char uchVal;
    int           nRetVal = 0;
    int           nData   = cmd.value();

    int nAdrStart = cmd.addr();
    int nAdrEnd   = nAdrStart + cmd.size() -1;

    for (int nAdr=nAdrStart; nAdr<=nAdrEnd; nAdr++)
    {
      uchCRCReg = 0;

      uchVal = 0xC0 | (nAdr & 0x0F);
      serial->WriteByte(uchVal);
      crc_byte(uchVal);

      uchVal = (nAdr >> 4) & 0x7F;
      serial->WriteByte(uchVal);
      crc_byte(uchVal);

      uchVal = ((nAdr & 0x7800) >> 10) | (nData & 0x01);
      serial->WriteByte(uchVal);
      crc_byte(uchVal);

      uchVal = (nData >> 1) & 0x7F;
      serial->WriteByte(uchVal);
      crc_byte(uchVal);

      crc_byte(0);
      serial->WriteByte(uchCRCReg);

      if (fReportOnWrite)
      {
        uchCRCReg = 0;
        
        nRetVal = serial->ReadByte();
        crc_byte((unsigned char)nRetVal);
        crc_byte((unsigned char)serial->ReadByte());
                                
        if (uchCRCReg != 0)
          setErr("CRC error (1)");

        results.push_back(nRetVal);
      }
    }
  }
  else
  {
    int nAdrStart = cmd.addr();
    int nAdrEnd   = nAdrStart + cmd.size() -1;

    for (int nAdr=nAdrStart; nAdr<=nAdrEnd; nAdr++)
    {
      unsigned char uchVal;
      int           nRetVal = 0;

      uchCRCReg = 0;

      uchVal = 0x80 | (nAdr & 0x0F);
      serial->WriteByte(uchVal);
      crc_byte(uchVal);
      
      uchVal = (nAdr >> 4) & 0x7F;
      serial->WriteByte(uchVal);
      crc_byte(uchVal);
      
      uchVal = (nAdr & 0x7800) >> 10;
      serial->WriteByte(uchVal);
      crc_byte(uchVal);
      crc_byte(0);
      serial->WriteByte(uchCRCReg);
                  
      uchCRCReg = 0;
      
      nRetVal = serial->ReadByte();
      crc_byte((unsigned char)nRetVal);
      crc_byte((unsigned char)serial->ReadByte());
            
      if (uchCRCReg != 0)
        setErr("CRC error (2)");

#if (VERBOSE_DBG != 0)
      std::cout << "empfangen: " << nRetVal << "\n";
#endif
  
      results.push_back(nRetVal);
    }
  }

  checkSerial();
}


avrdebug_ser::avrdebug_ser(const char* device,
                           int         baudrate,
                           std::string specialopts)
{
  setOK();
  
  serial = Serial::CreateSerial(device, baudrate, specialopts);
  
  checkSerial();
  
  // In diesem Modus startet die Firmware
  fReportOnWrite = true;
  // Die Adresse ist noch nicht bekannt
  nAdrFlags  = -1; 
  nAdrErrcnt = -1; 
}


avrdebug_ser::~avrdebug_ser() 
{
  delete serial;
}

int avrdebug_ser::clearErrCnt() 
{
  if (nAdrErrcnt < 0)
  {
    setErr("Adresse von ErrCnt ist nicht gesetzt!");
    return(-1);
  }
  
  DebugCmd    cmd;
  ResultBlock res;
    
  cmd.setWriteBlock(nAdrErrcnt, 1, 0);
  executeCmd(cmd, res);
  
  return(0);
}


int avrdebug_ser::getErrCnt() 
{
  if (nAdrErrcnt < 0)
  {
    setErr("Adresse von ErrCnt ist nicht gesetzt!");
    return(-1);
  }
  
  DebugCmd    cmd;
  ResultBlock res;
    
  cmd.setReadBlock(nAdrErrcnt, 1);
  res.clear();
  executeCmd(cmd, res);
  
  return(res[0]);
}


int avrdebug_ser::setReportOnWrite(bool fVal) 
{
  if (nAdrFlags < 0)
  {
    setErr("Adresse von Flags ist nicht gesetzt!");
    return(-1);
  }
  else
  {
    DebugCmd    cmd;
    ResultBlock res;
    
    // alte Flags lesen    
    cmd.setReadBlock(nAdrFlags, 1);
    res.clear();
    executeCmd(cmd, res);
    
    // Flags modifizieren
    if (fVal)
      res[0] &= 0xFF - 0x01;
    else
      res[0] |= 0x01;
    
    //
    fReportOnWrite = fVal;
    
    // neue Flags schreiben
    cmd.setWriteBlock(nAdrFlags, 1, res[0]);
    res.clear();
    executeCmd(cmd, res);
    
    return(0);
  }  
}

void avrdebug_ser::crc_byte(unsigned char uchData) 
{
  for (int n=0; n<8; n++)
  {
    uchCRCReg <<= 1;

    if ((uchData & 0x80) != 0)
      uchCRCReg |= 1;

    uchData <<= 1;

    if ((uchCRCReg & 0x80) != 0)
    {
      uchCRCReg ^= 0x31;
    }
  }
  uchCRCReg &= 0x7F;
}


void avrdebug_ser::checkSerial()
{
  if (!serial->IsOK())
    setErr(serial->GetFirstErr());
}
