// -*- mode: c; mode: fold -*-
//

/*
 * Copyright (C) 2005-2007  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "avrdebug_ser2.h"
#include "avrdebug_ser2_codes.h"
#include "config.h"
#include "lib_conversions.h"
#include "hirestime.h"

#include <iostream>

// Set to 256 to enable block read at its best, set to 1 to read single bytes.
// Any value in between works well, too.
#define MAX_BLOCKSIZE 256

// define to 0 to not use command blocks at all
// define to 1 to use command blocks internally
// define to 2 to split command blocks suitable for the interface to the adapter, too
// define to 3 to use command blocks on the interface to the adapter, too
#define USE_CMD_BLOCKS 3

#define CMD_BYTES 6

#define SHOW_SINGLE_ADAPTER_CMD 0

#define DEBUG_BLOCK 0

// todo: merker: set args -a ser2 -d /dev/ttyS1 -c 19200

avrdebug_ser2::avrdebug_ser2(const char* device,
                             int         baudrate,
                             eTargetCon  itargetCon,
                             int         itargetBaudrate,
                             std::string specialopts)
{
  setOK();

  targetCon       = itargetCon;
  targetBaudrate  = itargetBaudrate;
  
  serial = Serial::CreateSerial(device, baudrate, specialopts);
  
  checkSerial();
  
  sync();
}

int  avrdebug_ser2::setVoltages(float u_setpoint,
                                float u_max)            
{
  int nUNom = (int)(floor(0.5 + (256 * u_setpoint / REFVOLTAGE256)));
  int nUMax = (int)(floor(0.5 + (256 * u_max      / REFVOLTAGE256)));

  switch (targetCon)
  {
   case eTARGET_SPI:
    std::cerr << "Setting SPI voltage\n";
    writeCmd(CMD_debug_setvoltage, CMD_debug_setvoltage_spi, nUNom, nUMax, 3);
    break;
   case eTARGET_UART:
    std::cerr << "Setting UART voltage\n";
    writeCmd(CMD_debug_setvoltage, CMD_debug_setvoltage_uart, nUNom, nUMax, 3);
    break;
   case eTARGET_SELF:
    // nothing to do
    break;
   default:
    setErr("unknown target connection");
    break;
  }
  
  checkSerial();
  
  return(0);
}


int  avrdebug_ser2::resetTarget()                       
{
  writeCmd(CMD_debug_reset, 0, 0, 0, 3);

  return(0);
}


int  avrdebug_ser2::connect()                           
{
  switch (targetCon)
  {
    case eTARGET_SPI:
      {
        // 
        // The bit time in hundred nanoseconds is transferred:
        // 1*100ns..100*2^16ns
        // 10MHz..152Hz
        double dT = 1.0/(targetBaudrate*1.0E3);
        int    nT = (int)floor(0.5 + (dT/(100.0E-9)));
        
        return(writeCmd(CMD_debug_init, CMD_debug_init_spi, 
                        nT & 0xFF, (nT >> 8) & 0xFF, 3));
      }
      
    case eTARGET_UART:
      return(writeCmd(CMD_debug_init, CMD_debug_init_uart,
                      targetBaudrate & 0xFF, (targetBaudrate >> 8) & 0xFF, 3));
      
    case eTARGET_SELF:
      return(writeCmd(CMD_debug_init, CMD_debug_init_self, 0, 0, 3));
      
    default:
      setErr("unknown target connection");
      return(-1);
  }
}


int avrdebug_ser2::writeCmd(unsigned char b0,
                            unsigned char b1,
                            unsigned char b2,
                            unsigned char b3,
                            int           nCheck)
{
  int           nRetCode = 0;
  unsigned char uchVal;

  uchCRCReg = 0;

  uchVal = CMD_debug;
  crc_byte(uchVal);
  serial->WriteByte(uchVal);

  uchVal = b0;
  crc_byte(uchVal);
  serial->WriteByte(uchVal);

  uchVal = b1;
  crc_byte(uchVal);
  serial->WriteByte(uchVal);

  uchVal = b2;
  crc_byte(uchVal);
  serial->WriteByte(uchVal);

  uchVal = b3;
  crc_byte(uchVal);
  serial->WriteByte(uchVal);

  crc_byte(0);
  serial->WriteByte(uchCRCReg);
    
  if (nCheck > 0)
  {
    int reply = serial->ReadByte();
    if (nCheck > 1 && reply != b0)
    {
      setErr("Erstes Antwortbyte (" + itoStr(reply, ' ', 1) + ") war nicht das Kommando (" + itoStr(b0, ' ', 1));
    }
    if (nCheck > 2)
    {
      reply = serial->ReadByte();
      if (checkErr(reply, b0))
        nRetCode = -1;
    }
  }  
  checkSerial();
  return(nRetCode);
}


void avrdebug_ser2::executeCmd(DebugCmd      cmd,
                               ResultBlock&  results)   
{
  if (cmd.write())
  {
    int nData     = cmd.value();
    int nAdrStart = cmd.addr();
    int nAdrEnd   = nAdrStart + cmd.size() -1;
    int nRetVal;

    for (int nAdr=nAdrStart; nAdr<=nAdrEnd; nAdr++)
    {

#if SHOW_SINGLE_ADAPTER_CMD == 1
      std::cout << "write " << nData << " to addr " << nAdr << "\n";
#endif      

      writeCmd(CMD_debug_write, (nAdr & 0xFF), (nAdr >> 8) & 0xFF, nData & 0xFF, 0);

      results.push_back(serial->ReadByte());

      nRetVal = serial->ReadByte();
      checkErr(nRetVal);
    }
  }
  else
  {
    int nAdrStart = cmd.addr();
    int nSize     = cmd.size();
    int nRetVal;
    int bs;

    while (nSize)
    {
      if (nSize > MAX_BLOCKSIZE)
        bs = MAX_BLOCKSIZE;
      else
        bs = nSize;
      nSize -= bs;

#if SHOW_SINGLE_ADAPTER_CMD == 1
      std::cout << "read " << bs << " bytes from addr " << nAdrStart << "\n";
#endif

      writeCmd(CMD_debug_readblock, (nAdrStart & 0xFF), (nAdrStart >> 8) & 0xFF, bs, 0);

      for (int n=0; n<bs; n++)
      {
        results.push_back(serial->ReadByte());
        
#if SHOW_SINGLE_ADAPTER_CMD == 1
        std::cout << "    " << (int)(results[results.size()-1]) << "\n";
#endif

        nRetVal = serial->ReadByte();
        checkErr(nRetVal);
        nAdrStart++;
      }
    }
  }

  checkSerial();
}


int avrdebug_ser2::sync()                               
{
  unsigned char uchVal;
  int           retcode;

  // --- synchronize ------------------------------------------------
  uchVal = 0x1B;
  serial->WriteByte(uchVal);
  serial->WriteByte(uchVal);
  serial->WriteByte(uchVal);
  serial->WriteByte(uchVal);
  serial->WriteByte(uchVal);
  serial->WriteByte(uchVal);
  serial->WriteByte(uchVal);
  serial->WriteByte(uchVal);
  serial->WriteByte(uchVal);
  
  // --- Clear input buffer. Don't use serial->Flush() here! --------
  while (serial->ReadByte(uchVal) == 1)
  {
  }
  
  // --- read state, go idle ----------------------------------------
  {
    writeCmd(CMD_debug_get, CMD_debug_get_state, 0, 0, 0);
    int state = serial->ReadByte(true, 1000000);
    retcode   = serial->ReadByte(true,  500000);
    
    if (checkErr(retcode))
      return(-1);
    
    writeCmd(CMD_debug_get, CMD_debug_get_error, 0, 0, 0);
    int err   = serial->ReadByte(true, 500000);
    retcode   = serial->ReadByte(true, 500000);
    
    if (checkErr(retcode))
      return(-1);
    
    std::cerr << "Adapter state = " << state << "\n"
      << "Adapter error = " << err << "\n";
    
    switch (state)
    {
      case 0:
        // idle
        break;
        
      case 1:
        {
          std::cerr << "Adapter in programming mode. Leave.\n";
          
          unsigned char uchVal;
          
          uchVal = 'L';
          serial->WriteByte(uchVal);
          retcode = serial->ReadByte(true, 500000);
          if (checkErr(retcode))
            return(-1);
        }
        break;
        
      case 2:
        std::cerr << "Adapter in debug mode. Leave.\n";
        disconnect();
        break;
        
      default:
        // unknown mode.
        setErr("unknown state of adapter");
        return(-1);
        break;
    }
  }
  std::cerr << "Adapter should be idle now.\n";
  
  
  // --- Read receive buffer size -----------------------------------
  {
    int nData;
    
    writeCmd(CMD_debug_get, CMD_debug_get_inbufsize, 0, 0, 0);
    nData   = serial->ReadByte(true, 500000);
    retcode = serial->ReadByte(true, 500000);
    if (checkErr(retcode))
      return(-1);
    // one command is CMD_BYTES bytes
    uMaxCmdsInInternalBlock = nData / CMD_BYTES;
    std::cerr << "Receive buffer of adapter is " << nData << " Bytes ("
      << uMaxCmdsInInternalBlock << " commands)\n";
  }
  
  // --- read voltages ----------------------------------------------
  {
    int nData;

    std::cerr.precision(3);

    writeCmd(CMD_debug_get, CMD_debug_get_voltage, CMD_debug_get_voltage_spi, CMD_debug_get_voltage_spi_setpoint);
    nData   = serial->ReadByte();
    retcode = serial->ReadByte();
    if (checkErr(retcode))
      return(-1);
    std::cerr << "  Voltage SPI setpoint  = " << (REFVOLTAGE256*nData/256) << " V\n";

    writeCmd(CMD_debug_get, CMD_debug_get_voltage, CMD_debug_get_voltage_spi, CMD_debug_get_voltage_spi_max);
    nData   = serial->ReadByte();
    retcode = serial->ReadByte();
    if (checkErr(retcode))
      return(-1);
    std::cerr << "  Voltage SPI max       = " << (REFVOLTAGE256*nData/256) << " V\n";

    writeCmd(CMD_debug_get, CMD_debug_get_voltage, CMD_debug_get_voltage_spi, CMD_debug_get_voltage_spi_cur);
    nData   = serial->ReadByte();
    retcode = serial->ReadByte();
    if (checkErr(retcode))
      return(-1);
    std::cerr << "  Voltage SPI current   = " << (REFVOLTAGE256*nData/256) << " V\n";

    writeCmd(CMD_debug_get, CMD_debug_get_voltage, CMD_debug_get_voltage_uart, CMD_debug_get_voltage_uart_setpoint);
    nData   = serial->ReadByte();
    retcode = serial->ReadByte();
    if (checkErr(retcode))
      return(-1);
    std::cerr << "  Voltage UART setpoint = " << (REFVOLTAGE256*nData/256) << " V\n";

    writeCmd(CMD_debug_get, CMD_debug_get_voltage, CMD_debug_get_voltage_uart, CMD_debug_get_voltage_uart_max);
    nData   = serial->ReadByte();
    retcode = serial->ReadByte();
    if (checkErr(retcode))
      return(-1);
    std::cerr << "  Voltage UART max      = " << (REFVOLTAGE256*nData/256) << " V\n";

    writeCmd(CMD_debug_get, CMD_debug_get_voltage, CMD_debug_get_voltage_uart, CMD_debug_get_voltage_uart_cur);
    nData   = serial->ReadByte();
    retcode = serial->ReadByte();
    if (checkErr(retcode))
      return(-1);
    std::cerr << "  Voltage UART current  = " << (REFVOLTAGE256*nData/256) << " V\n";

  }
  
  checkSerial();

  return(0);
}

void avrdebug_ser2::disconnect()
{
  // std::cout << "void avrdebug_ser2::disconnect()\n";
  writeCmd(CMD_debug_exit, 0, 0, 0, 3);
}

     avrdebug_ser2::~avrdebug_ser2()                    
{
  std::cout << "serial device aufräumen:\n";
  delete serial;
}

int  avrdebug_ser2::checkErr(int nVal, int nCmd)
{
  std::string msg = "";
  
  if (nCmd >= 0)
  {
    char charcmd = (char)nCmd;
    msg = std::string(" (cmd=") + charcmd + std::string(")");
  }
  
  switch (nVal)
  {
   case REPLY_ok:
    return(0);
    break;
   case REPLY_error:
    setErr("error connecting to target" + msg);
    break;
   case REPLY_unknown:
    setErr("target adapter does not understand me" + msg);
    break;
   case REPLY_errCRC:
    setErr("target adapter found CRC error" + msg);
    break;
   default:
    setErr("unexpected reply from target adapter" + msg);
    break;
  }
  return(-1);
}


void avrdebug_ser2::crc_byte(unsigned char uchData)     
{
  bool fXOR;

  for (int n=0; n<8; n++)
  {
    fXOR = ((uchCRCReg & 0x80) != 0);

    uchCRCReg <<= 1;

    if ((uchData & 0x80) != 0)
      uchCRCReg |= 1;

    uchData <<= 1;

    if (fXOR)
      uchCRCReg ^= 0x31;
  }
}


bool avrdebug_ser2::addCmd(DebugCmd cmd)                
{
  unsigned int size = cmdblock.size();

#if (DEBUG_BLOCK > 0 || GLOB_DEBUG > 1)
  std::cout << "bool avrdebug_ser2::addCmd(" << cmd.describe() << ")\n";  
#endif
  
  // Can this cmd be added to a previous read block?
  // Write blocks are not combined, as the adapter can
  // only write single bytes in one command.
  if (!cmd.write() && size)
  {
    if (cmdblock[size-1].combine(&cmd))
    {
#if DEBUG_BLOCK > 0
      std::cout << "return: command combined\n";
#endif
      fExecuted = false;
      return(true);
    }
  }

#if USE_CMD_BLOCKS > 0
  
  // todo: hier in diesem Abschnitt gibt es bereits Fehler.
  
  // How many internal commands does cmd equal?
  unsigned int uNrCmds = nrOfInternalCmds(cmd);
    
  if (uCmdsInExternalBlock == 0 || uCmdsInExternalBlock + uNrCmds <= uMaxCmdsInInternalBlock)
  {
    uCmdsInExternalBlock += uNrCmds;
    cmdblock.push_back(cmd);
    fExecuted = false;
#if DEBUG_BLOCK > 0
    std::cout << "return: command added\n";
#endif
    return(true);
  }
  else
  {
#if DEBUG_BLOCK > 0
    std::cout << "return: command refused\n";
#endif
    return(false);
  }
#else  
  // no further optimization
  if (size)
    return(false);
  else
  {
    cmdblock.push_back(cmd);
    fExecuted = false;
    return(true);
  }
#endif  
}


void avrdebug_ser2::clearCmdBlock()                     
{
#if DEBUG_BLOCK > 0
  std::cout << "void avrdebug_ser2::clearCmdBlock()\n";  
#endif
  avrdebug::clearCmdBlock();
  uCmdsInExternalBlock = 0;
}


void avrdebug_ser2::executeBlock(std::vector<DebugCmd> cmds,
                                 ResultBlock&          results) 
{
  
#if DEBUG_BLOCK > 0
  std::cout << "void avrdebug_ser2::executeBlock(...)\n";
#endif
  
  unsigned int uNrOfCmds = cmds.size();
  
#if USE_CMD_BLOCKS < 3
  // execute block as single commands
  
  for (unsigned int n=0; n<uNrOfCmds; n++)
  {
    // std::cout << cmds[n].describe() << "\n";
    executeCmd(cmds[n], results);
  }      
#else
  // execute block as block
  unsigned int receivebytes = 0;
  unsigned int adr;
  unsigned int size;
  int          data;

  // set size of block
#if SHOW_SINGLE_ADAPTER_CMD  == 1
  std::cout << "wait for " << uNrOfCmds*CMD_BYTES << " bytes\n";
#endif
  
  // Möglicherweise kommt jetzt gar kein Block:
  if (uNrOfCmds > 1)
    writeCmd(CMD_debug_waitbytes, uNrOfCmds*CMD_BYTES, 0, 0);
  
  // send commands
  for (unsigned int n=0; n<uNrOfCmds; n++)
  {
    size = cmds[n].size();
    adr  = cmds[n].addr();
    
    receivebytes += size;
    
    if (cmds[n].write())
    {
      data = cmds[n].value();
      writeCmd(CMD_debug_write, (adr & 0xFF), (adr >> 8) & 0xFF, data & 0xFF);      
#if SHOW_SINGLE_ADAPTER_CMD == 1
      std::cout << "write " << data << " to addr " << adr << "\n";
#endif      
    }
    else
    {
      writeCmd(CMD_debug_readblock, (adr & 0xFF), (adr >> 8) & 0xFF, size & 0xFF);
#if SHOW_SINGLE_ADAPTER_CMD == 1
      std::cout << "read " << size << " bytes from addr " << adr << "\n";
#endif
    }
  }
  
  // receive replies
#if SHOW_SINGLE_ADAPTER_CMD == 1
  std::cout << "receiving " << receivebytes << " replies\n";
#endif
  for (unsigned int n=0; n<receivebytes; n++)
  {
    results.push_back(serial->ReadByte());

    data = serial->ReadByte();
    checkErr(data);    
  }
#endif
  checkSerial();
}


void avrdebug_ser2::executeCmdBlock(ResultBlock& results)       
{
  
#if DEBUG_BLOCK > 0
  std::cout << "void avrdebug_ser2::executeCmdBlock(ResultBlock& results)\n";
#endif
  
#if USE_CMD_BLOCKS < 2
  avrdebug::executeCmdBlock(results);
#else
  // Split the external command block into blocks which can
  // be handled by the adapter.
  unsigned int          nExternalCmdBlockSize = cmdblock.size();
  std::vector<DebugCmd> internalCmds;
  DebugCmd              cmd;
  unsigned int          uSize;
  unsigned int          bs;
  unsigned int          maxbs;
  
  for (unsigned int n=0; n<nExternalCmdBlockSize; n++)
  {
    cmd   = cmdblock[n];
    uSize = cmd.size();

    if (cmd.write())
      maxbs = 1;
    else
      maxbs = MAX_BLOCKSIZE;
    
    while (uSize)
    {
      if (uSize > maxbs)
        bs = maxbs;
      else
        bs = uSize;
      uSize -= bs;

      cmd.setSize(bs);
      internalCmds.push_back(cmd);
      cmd.setAddr(cmd.addr() + bs);
            
      if (internalCmds.size() >= uMaxCmdsInInternalBlock)
      {
        executeBlock(internalCmds, results);
        internalCmds.clear();
      }
      
    }    
  }
  
  // execute remaining commands
  if (internalCmds.size())      
    executeBlock(internalCmds, results);
  
  fExecuted = true;
#endif
}


unsigned int avrdebug_ser2::nrOfInternalCmds(DebugCmd cmd)      
{
  unsigned int uNrCmds;
  
  if (cmd.write())
    uNrCmds = cmd.size();
  else  
  {
    uNrCmds = cmd.size() / MAX_BLOCKSIZE;
    if (cmd.size() % MAX_BLOCKSIZE > 0)
      uNrCmds++;
  }

  return(uNrCmds);
}


void avrdebug_ser2::checkSerial()
{
  if (!serial->IsOK())
    setErr(serial->GetFirstErr());  
}
