// -*- mode: c; mode: fold -*-
//

/*
 * Copyright (C) 2005-2009  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include <iostream>
#include <fstream>
#include <string>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef USE_READLINE
# include <stdio.h>
# include <readline/readline.h>
# include <readline/history.h>
#endif

#ifdef HANDLE_CTRL_C
# include <signal.h>
#endif

#include "lib_conversions.h"
#include "usrcmd.h"
#include "avrdebug.h"
#include "avrdebug_par.h"
#include "avrdebug_ser.h"
#include "avrdebug_ser2.h"
#include "../version.h"
#include "config.h"

// man console_codes
// #define COLORNORMAL cout << "\x1b[1m\x1b[37m\x1b[40m"
#define COLORNORMAL std::cout << "\x1b[0m"
//#define COLORDEBUG  cout << "\x1b[30m\x1b[41m"
#define COLORDEBUG  std::cout << "\x1b[7m"

struct OutputOptions
{
  std::string   buf;
  bool          fTableOutput;
  int           nTableRow;
  std::ostream* out;
};

#ifdef HANDLE_CTRL_C
static bool fEndRequested = false;

/**
 * Handle CTRL-C and others
 */
void signal_handler(int signum)
{
  std::cerr << "Received signal: " << signum << "  " << strsignal(signum) << "\n";
  fEndRequested = true;
}
#endif

void showHelp()
{
  std::cerr << "Usage:\n"
    " avrdebug [-a adapter] [-d addressOrDevice] [-c adapterRateOrClock]\n"
    "          [-i targetInterface] [-C targetRateOrClock]\n"
    "          [-u voltagesetpoint-max] \n"
    "          [-r] [-t]\n"
    "          [-f filename]\n"
    "          [-s SpecialOptionString]\n"
    "          [read command]\n"
    "\n"
    "  adapter:            par, ser, ser2\n"
    "  addressOrDevice:    par: address of parallel port (default: 888)\n"
    "                      ser, ser2: serial device (/dev/ttyS1)\n"
    "  adapterRateOrClock: baudrate for serial connection to adapter\n"
    "  targetInterface:    SPI, UART or SELF\n"
    "  targetRateOrClock:  baudrate from adapter to target (UART),\n"
    "                      clock in kHz (SPI)\n"
    "  voltagesetpoint-max for example 3.2-3.5\n"
    "\n"
    "  -r                  Resets the target if possible\n"
    "  -t                  Output can be used with gnuplot\n"
    "\n"
    "  -f filename         Read commands from file\n"
    "\n"
    "  read command        a read command will be looped\n";

  std::cerr << "\n";
}

void errMissingArgAfter(char arg)
{
  std::cerr << "Missing argument after -" << arg << "\n\n";
  showHelp();
  throw std::exception();
}

void explainCmd(std::string descr)
{
  std::cout << "   " << descr << "\n";
}

long long unsigned int getValue(const unsigned char* results,
                                UserCmd              cmd)
{
  long long unsigned int uVal = 0;
  int                    nWidthBits  = cmd.width();
  int                    nWidthBytes = nWidthBits / 8;
  
  if ((nWidthBits % 8) != 0)
  {
    std::cerr << "only multiples of bytes allowed: " << nWidthBits << "\n";
    throw std::exception();
  }

  for (int n=0; n<nWidthBytes; n++)
    uVal |= ((long long unsigned int)(results[n])) << (n*8);
  
  return(uVal);
}

std::string getDisplayString(const unsigned char* results,
                             UserCmd              cmd,
                             unsigned int         uAdr,
                             OutputOptions        outopt,
                             bool&                fExplained)
{
  std::string         line;
  std::string         characters = "";
  std::string         address    = "[";
  long long int          nVal;
  long long unsigned int uVal = 0;
  int                    nWidthBits  = cmd.width();
  int                    nWidthBytes = nWidthBits / 8;
  
  if ((nWidthBits % 8) != 0)
  {
    std::cerr << "only multiples of bytes allowed: " << nWidthBits << "\n";
    throw std::exception();
  }
  
  uVal = getValue(results, cmd);
  
  if ((uVal & (1LL<<(nWidthBits-1))) != 0)
  {
    nVal = (long long int)(uVal) - (1LL<<nWidthBits);
  }
  else
  {
    nVal = uVal;
  }
  
  for (int n=nWidthBytes-1; n>=0; n--)
  {
    int thisbyte = (uVal>>(n*8)) & 0xFF;
    address += itoStr(uAdr+n, ' ', 5);
    if (n>0)
      address += ":";
    if (isprint( thisbyte  ))
      characters.push_back( (char)(thisbyte) );
    else
      characters.push_back(' ');
  }
  address += "]";
  
  switch (cmd.display())
  {
   case UserCmd::eSIGNED:
    line = itoStr(nVal, ' ', 6) + " ";
    break;
   case UserCmd::eUNSIGNED:
    
    line = itoStr(uVal, ' ', 6) + " ";
    
    // todo nur zum Test:
    // line += " spezial: " + itoStr(uVal >> 7, ' ', 6) + ": " + ftoStr((uVal & 0x7F)/12.0, 1, 1) + " ";
    
    break;

   case UserCmd::eHEX:
      line = "  " + itoHexStr(uVal, cmd.width()) + " ";    
    break;
        
   default:
    line  = itoStr(nVal, ' ', 6) + " ";
    line += itoStr(uVal, ' ', 6) + " ";
    line += characters;
    break;
  }

  if (cmd.showValue() == UserCmd::eNORMAL)
  {
    if (!outopt.fTableOutput)
    {
      line = "  " + address + " " + line;
            
      if (!fExplained)
      {
        line += "   " + cmd.explainCmd();
        fExplained = true;
      }     
    }
    return(line);
  }
  
  return("");
}

void executeCmdBlock(avrdebug*              avr,
                     std::vector<UserCmd>&  cmds,
                     OutputOptions&         outopt)
{
  ResultBlock         results;
  unsigned int        uNrOfUserCmds = cmds.size();
  int                 nWidth;
  bool                fWrite;
  UserCmd::eDisplay   display;
  unsigned int        uStartAdr;
  unsigned int        uEndAdr;
  unsigned int        uAdr;
  bool                fExplained;
  std::string         line;
  std::string         characters;
  std::string         address;

  unsigned int        uByte = 0;

  results.clear();
  
  avr->executeCmdBlock(results);
  avr->clearCmdBlock();

  for (unsigned int n=0; n<uNrOfUserCmds; n++)
  {
    nWidth     = cmds[n].width();
    fWrite     = cmds[n].write();
    display    = cmds[n].display();
    uStartAdr  = cmds[n].adr();
    uEndAdr    = uStartAdr + cmds[n].size();
    fExplained = false;
    
#if (GLOB_DEBUG > 2)
    std::cout << "User command:\n";
    std::cout << "   nWidth     :" << nWidth     << "\n";
    std::cout << "   fWrite     :" << fWrite     << "\n";
    std::cout << "   display    :" << display    << "\n";
    std::cout << "   uStartAdr  :" << uStartAdr  << "\n";
    std::cout << "   uEndAdr    :" << uEndAdr    << "\n";
    std::cout << "   fExplained :" << fExplained << "\n";
#endif

    if (cmds[n].cmd() == UserCmd::eNEWLINE)
    {
      if (outopt.buf.length())
        *(outopt.out) << outopt.nTableRow++ << " " << outopt.buf << "\n";
      outopt.buf = "";
    }
    else if (cmds[n].cmd() == UserCmd::eECHO)
    {
      *(outopt.out) << cmds[n].getEcho() << "\n";
    }
    else
    {
      for (uAdr=uStartAdr; uAdr<uEndAdr; )
      {
        if (uByte >= results.size())
        {
          std::cerr << "unexpected error: no more results\n";
        }
        
        outopt.buf += getDisplayString(&(results[uByte]), cmds[n], uAdr, outopt, fExplained);
      
        if (!outopt.fTableOutput)
        {
          *(outopt.out) << outopt.buf << "\n";
          outopt.buf = "";
        }
                
        uAdr  += nWidth/8;
        uByte += nWidth/8;
      }
    }
  }

  cmds.clear();
}

void showHelpInteractive()
{
  std::cerr << "Commands in interactive mode:\n"
    "   show help:            h\n"
    "   read byte(s):         r  A [-B],     r  A [*D]\n"
    "   read word(s):         rw A [-B],     rw A [*D]\n"
    "   write byte(s):        w  A [-B] : C, w  A [*D] : C\n"
    "   wait until:           u A==C, u A!=C, u A>C, u A<C\n"
    "   quit:                 q\n"
    "   read file:            f FILENAME\n";
  
  std::cerr << "Every address (A, B) and number (C, D) needs to be written decimal.\n\n";
  std::cerr << "Optional modifiers (u, s, h, q, Q) are allowed after r, rw, w, u. They indicate\n"
               "how to display the variables content: unsigned decimal, signed decimal,\n"
               "hexadecimal. q and Q suppress output. With u (wait until), q means to only\n"
               "display the value when it matched, Q never displays the value read.\n\n";
  std::cerr << "To indicate variable length, you may also use (8, 16, 24, 32) with r (read)\n"
               "and u (wait until), but you have to use one of the optional modifiers in this\n"
               "case, for example 'r24u A' or 'u16 A == C'\n\n"
               "This help is not complete, please take a look at the manual!\n";
}

std::string getline(std::ifstream& in)
{
  char Zeile[500];
  in.getline(Zeile, 499);
  std::string befehlEingabe = Zeile;
  return(befehlEingabe);
}

int main(int argc, char* argv[])
{
  enum eInputMode { eINTERACTIVE, eCMDLINE, eFILE };
  eInputMode            inputmode;
  UserCmd               uc;
  std::vector<UserCmd>  cmds;
  UserVars*             vars = new UserVars();

  int            nArgCmd=0;
  std::ifstream  infile;
  std::ofstream* outfile = 0;
  bool           fReset       = false;
  std::string    cmdfile      = "";
  std::string    specialopts  = "";
  avrdebug*      avr          = (avrdebug*)0;
  OutputOptions  outopt;
  int            nRetCode     = 0;
  bool           fFileContainsLoopBegin = false;

  uc.SetVarCache(vars);

  std::cerr << "avrdebug version " << AVRDEBUGVERSION << "\n";

  outopt.fTableOutput = false;
  outopt.nTableRow    = 0;
  
  try
  {
    // Argumente parsen
    {
      int           nArgC              = 1;
      std::string   adapter            = "par";
      std::string   addressOrDevice    = "888";
      std::string   adapterRateOrClock = "19200";
      std::string   targetRateOrClock  = "19200";
      std::string   targetInterface    = "SPI";
      std::string   outfilename        = "";
      std::string   voltage            = "";
      
      while (nArgC < argc)
      {
        
        // Befehl?
        if (nArgC+1 == argc)
        {
          if (argv[nArgC][0] == 'r')
          {
            nArgCmd = nArgC;
          }
        }
        
        if (argv[nArgC][0] == '-')
        {
          switch (argv[nArgC][1])
          {
            case 'h':
              showHelp();
              return(0);
              break;
              
            case 'a':
              if (nArgC+1 < argc)
              {
                nArgC++;
                adapter = argv[nArgC];
              }
              else
                errMissingArgAfter(argv[nArgC][1]);
              break;
              
            case 'r':
              fReset = true;
              break;
              
            case 'd':
              if (nArgC+1 < argc)
              {
                nArgC++;
                addressOrDevice = argv[nArgC];
              }
              else
                errMissingArgAfter(argv[nArgC][1]);
              break;
              
            case 's':
              if (nArgC+1 < argc)
              {
                nArgC++;
                // surround by ',' for easier parsing.
                specialopts = std::string(",") + argv[nArgC] + std::string(",");
                // remove whitespace (values excluded)
                std::string tmp;
                bool fIn = false;
                for (std::string::size_type n=0; n<specialopts.length(); n++)
                {
                  if (fIn == false)
                  {
                    if (specialopts[n] != ' ')
                    {
                      tmp.push_back(specialopts[n]);
                      if (specialopts[n] == '=')
                        fIn = true;
                    }
                  }
                  else
                  {
                    tmp.push_back(specialopts[n]);
                    if (specialopts[n] == ',')
                      fIn = false;
                  }                                
                }
                specialopts = tmp;
              }
              else
                errMissingArgAfter(argv[nArgC][1]);
              break;
              
            case 'u':
              if (nArgC+1 < argc)
              {
                nArgC++;
                voltage = argv[nArgC];
              }
              else
                errMissingArgAfter(argv[nArgC][1]);
              break;
              
            case 'f':
              if (nArgC+1 < argc)
              {
                nArgC++;
                cmdfile = argv[nArgC];
              }
              else
                errMissingArgAfter(argv[nArgC][1]);
              break;
              
            case 'c':
              if (nArgC+1 < argc)
              {
                nArgC++;
                adapterRateOrClock = argv[nArgC];
              }
              else
                errMissingArgAfter(argv[nArgC][1]);
              break;
              
            case 'C':
              if (nArgC+1 < argc)
              {
                nArgC++;
                targetRateOrClock = argv[nArgC];
              }
              else
                errMissingArgAfter(argv[nArgC][1]);
              break;
              
            case 'i':
              if (nArgC+1 < argc)
              {
                nArgC++;
                targetInterface = argv[nArgC];
              }
              else
                errMissingArgAfter(argv[nArgC][1]);
              break;
              
            case 'o':
              if (nArgC+1 < argc)
              {
                nArgC++;
                outfilename = argv[nArgC];
              }
              else
                errMissingArgAfter(argv[nArgC][1]);
              break;
              
            case 't':
              outopt.fTableOutput = true;
              break;
              
            default:
              std::cerr << "unknown option: " << argv[nArgC] << "\n";
              showHelp();
              return(-1);
              break;
          }
        }
        
        // nächstes Argument
        nArgC++;
      }
      
      std::cerr << "Options:\n"
        << "  fTableOutput = " << outopt.fTableOutput << "\n"
        << "  fReset       = " << fReset << "\n"
        << "  adapter      = " << adapter << "\n"
        << "  addressOrDevice    = " << addressOrDevice << "\n"
        << "  adapterRateOrClock = " << adapterRateOrClock << "\n"
        << "  targetRateOrClock  = " << targetRateOrClock << "\n"
        << "  targetInterface    = " << targetInterface << "\n"
        << "  voltage            = " << voltage << "\n"
        << "  cmdfile            = " << cmdfile << "\n"
        << "  outfile            = " << outfilename << "\n";
      
      if (adapter.compare("par") == 0)
      {
        avr = new avrdebug_par(atoi(addressOrDevice.c_str()));
      }
      else if (adapter.compare("ser") == 0)
      {
        avr = new avrdebug_ser(addressOrDevice.c_str(), atoi(adapterRateOrClock.c_str()), specialopts);
      }
      else if (adapter.compare("ser2") == 0)
      {
        enum avrdebug_ser2::eTargetCon con;
        
        if (targetInterface.compare("UART") == 0)
          con = avrdebug_ser2::eTARGET_UART;
        else if (targetInterface.compare("SPI") == 0)
          con = avrdebug_ser2::eTARGET_SPI;
        else
          con = avrdebug_ser2::eTARGET_SELF;
        
        avr = new avrdebug_ser2(addressOrDevice.c_str(), atoi(adapterRateOrClock.c_str()),
                                con, atoi(targetRateOrClock.c_str()), specialopts);
      }
      else
      {
        std::cerr << "unknown adapter: " << adapter << "\n";
        showHelp();
        return(-1);
      }
      
      if (!avr->isOK())
      {
        std::cerr << avr->getErr() << "\n";
        throw std::exception();
      }
      
      if (fReset)
      {
        avr->resetTarget();
        if (!avr->isOK())
        {
          std::cerr << avr->getErr() << "\n";
          throw std::exception();
        }
      }
      
      if (voltage.length() > 0)
      {
        float                  unom, umax;
        std::string::size_type pos         = voltage.find("-");
        
        if (pos > 0 && pos < voltage.length()-1)
        {
          unom = atof(voltage.substr(0, pos).c_str());
          umax = atof(voltage.substr(pos+1).c_str());
          avr->setVoltages(unom, umax);
          
          if (!avr->isOK())
          {
            std::cerr << avr->getErr() << "\n";
            throw std::exception();
          }
          delete avr;
          return(0);
        }
        else
        {
          showHelp();
          throw std::exception();
        }
      }
      
      avr->connect();
      
      if (!avr->isOK())
      {
        std::cerr << avr->getErr() << "\n";
        avr->disconnect();
        throw std::exception();
      }
      avr->clearCmdBlock();

      if (outfilename.length() > 0)
      {
        outfile = new std::ofstream();
        outfile->open(outfilename.c_str());
        outopt.out = outfile;
        if (!(*outfile))
        {
          std::cerr << "Konnte die Datei " << outfilename << " nicht öffnen.\n";
          outopt.out = &std::cout;
        }
      }
      else
        outopt.out = &std::cout;
    }
    
    if (cmdfile.length() > 0)
    {
      infile.open(cmdfile.c_str());
      if (!infile)
      {
        std::cerr << "Konnte die Datei " << cmdfile << " nicht öffnen.\n";
        cmdfile = "";
      }
    }
        
    std::cerr << "\n";
    
    showHelpInteractive();
    
#ifdef HANDLE_CTRL_C
    // Handle CTRL-C and others
    signal(SIGTERM, signal_handler);
    signal(SIGINT,  signal_handler);
    signal(SIGQUIT, signal_handler);
#endif
    
    do
    {
      // Read a command into 'uc'      
      {
        std::string    befehlEingabe;
        
        if (cmdfile.length() > 0)
        {
          bool fAgain;
          do
          {
            // Befehle aus Datei lesen
            inputmode = eFILE;
            befehlEingabe = getline(infile);
            
            if (!infile)
            {
              befehlEingabe = "eof";
              std::cerr << "   warning: file did not end with \"eof\" or \"loop\"!\n";
            }
            
            fAgain = (trim(befehlEingabe).find("loop_begin:") == 0);
            if (fAgain)
            {
              fFileContainsLoopBegin = true;
            }            
          }
          while (fAgain);
        }
        else if (nArgCmd != 0 && argv[nArgCmd][0] == 'r')
        {
          inputmode     = eCMDLINE;
          befehlEingabe = argv[nArgCmd];
        }
        else
        {
          inputmode = eINTERACTIVE;
#ifdef USE_READLINE
          char* line = readline("Command: ");
          befehlEingabe = line;
          if (line != (char*)0)
          {
            if (befehlEingabe.length() > 0)
              add_history(line);
            free(line);
          }
          else
            befehlEingabe = "";
#else
          COLORNORMAL << "Command: ";
          std::cin >> befehlEingabe;
#endif
        }
        
        uc.setCmd(befehlEingabe);
      }
      
      // handle command                
      if (uc.isOK())
      {
        switch (uc.cmd())
        {
          case UserCmd::eRW:       
            {
              DebugCmd     cmd;
              
              if (uc.write())
              {
                if (uc.width() != 8)
                {
                  std::cerr << "it is not possible to write something else than bytes...\n";
                }
                cmd.setWriteBlock(uc.adr(), uc.size(), uc.value());
              }
              else
                cmd.setReadBlock( uc.adr(), uc.size());
              
              if (!avr->addCmd(cmd))
              {
                executeCmdBlock(avr, cmds, outopt);
                
                avr->addCmd(cmd);
              }
              cmds.push_back(uc);
              
              if (inputmode == eINTERACTIVE)
              {
                executeCmdBlock(avr, cmds, outopt);
              }
            }
            break;
            
          case UserCmd::eSLEEP:
            {
              // flush cmd block if necessary
              if (!avr->executed() || cmds.size())
              {
                executeCmdBlock(avr, cmds, outopt);
                avr->clearCmdBlock();
              }
              
              if (!outopt.fTableOutput && uc.showValue() == UserCmd::eNORMAL)
                explainCmd(uc.explainCmd());
              
              usleep(uc.value());
            }
            break;
            
          case UserCmd::eWAIT:     
            {
              // flush cmd block if necessary
              if (!avr->executed() || cmds.size())
              {
                executeCmdBlock(avr, cmds, outopt);
              }
              
              if (!outopt.fTableOutput && uc.showValue() == UserCmd::eNORMAL)
                explainCmd(uc.explainCmd());
              
              DebugCmd      cmd;
              ResultBlock   results;
              bool          fDone      = false;
              bool          fExplained = true;
              unsigned int  uVal;
              
              cmd.setReadBlock(uc.adr(), uc.size());
              if (avr->addCmd(cmd) == false)
              {
                std::cerr << "unexpected error!\n";
              }
              
              do
              {
                results.clear();
                avr->executeCmdBlock(results);
                
                uVal = getValue(&(results[0]), uc);
                
                switch (uc.relation())
                {
                  case UserCmd::eEQUAL:
                    fDone = (uVal == (unsigned int)uc.value());
                    break;
                  case UserCmd::eNOTEQUAL:
                    fDone = (uVal != (unsigned int)uc.value());
                    break;
                  case UserCmd::eGREATER:
                    fDone = (uVal > (unsigned int)uc.value());
                    break;
                  case UserCmd::eSMALLER:
                    fDone = (uVal < (unsigned int)uc.value());
                    break;
                }
                
                if (fDone)
                  uc.WriteValueToVarCache(uVal);
                
                if (uc.showValue() == UserCmd::eNORMAL)
                {
                  // show value
                  outopt.buf += getDisplayString(&(results[0]), uc, uc.adr(), outopt, fExplained);
                  
                  if (!outopt.fTableOutput)
                  {
                    *(outopt.out) << outopt.buf << "\n";
                    outopt.buf = "";
                  }
                }
              }
              while (!fEndRequested && !fDone && avr->isOK());
              
              // my read cmd is still there:
              avr->clearCmdBlock();
              
              if (uc.showValue() == UserCmd::eQUIET && fDone)
              {
                uc.setShowValue(UserCmd::eNORMAL);
                fExplained = false;
                // show value
                uc.setShowValue(UserCmd::eNORMAL);
                outopt.buf += getDisplayString(&(results[0]), uc, uc.adr(), outopt, fExplained);
                
                if (!outopt.fTableOutput)
                {
                  *(outopt.out) << outopt.buf << "\n";
                  outopt.buf = "";
                }
              }
              
            }
            break;
            
          case UserCmd::eEOF:      
            {
              if (cmdfile.length() > 0)
              {
                cmdfile = "";
                infile.close();
                
                // flush cmd block if necessary
                if (!avr->executed() || cmds.size())
                {
                  executeCmdBlock(avr, cmds, outopt);
                }
                
                if (!outopt.fTableOutput)
                  explainCmd(uc.explainCmd());
              }
              else
              {
                std::cerr << "   error: not reading from file!\n";
              }
            }
            break;
            
          case UserCmd::eLOOP:     
            {
              if (!outopt.fTableOutput)
                explainCmd(uc.explainCmd());
              
              if (cmdfile.length() > 0)
              {
                infile.close();
                infile.open(cmdfile.c_str());
                if (fFileContainsLoopBegin)
                {
                  while (trim(getline(infile)).find("loop_begin:") != 0)
                  {
                  }
                }
              }
              else
              {
                std::cerr << "   error: not reading from file!\n";
              }
            }
            break;
            
          case UserCmd::eREADFILE: 
            {
              explainCmd(uc.explainCmd());
              if (cmdfile.length() == 0)
              {
                cmdfile = uc.getFilename();
                infile.clear(); // in case of a bad state before
                infile.open(cmdfile.c_str());
                
                if (!infile)
                {
                  std::cerr << "   Konnte die Datei " << cmdfile << " nicht öffnen.\n";
                  cmdfile = "";
                }
              }
              else
              {
                std::cerr << "   error: not allowed when reading from file!\n";
              }
            }
            break;
            
          case UserCmd::eNEWLINE:  
            {
              if (outopt.fTableOutput)
                cmds.push_back(uc);
            }
            break;
            
          case UserCmd::eHELP:     
            showHelpInteractive();
            break;
            
          case UserCmd::eECHO:     
            cmds.push_back(uc);
            break;
            
          default:
            break;
        }
      }
      else
        std::cout << "   error: " << uc.getErr() << "\n";
      
    }
    while (uc.cmd() != UserCmd::eQUIT && avr->isOK() && (inputmode == eINTERACTIVE || uc.isOK())
#ifdef HANDLE_CTRL_C
           && fEndRequested == false
#endif
           );
    
    // flush cmd block if necessary
    if (!avr->executed() || cmds.size())
    {
      executeCmdBlock(avr, cmds, outopt);
    }
    
    avr->disconnect();
    
    if (!avr->isOK())
    {
      nRetCode = -1;
      std::cerr << avr->getErr() << "\n";
    }
  }
  catch (std::exception& e)
  {
    std::cerr << "Exception: " << e.what() << "\n";
    nRetCode = -1;
  }

  if (avr != 0)
    delete avr;
  
  if (outfile)  
    outfile->close();
  
  return(nRetCode);
}
