/*
 * Copyright (C) 2007  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "serial_ftdi.h"

// 0: keine Ausgaben
// 1: nur Fehler
// 2: weiteres
#define DEBUG 2

#if DEBUG > 0
# include <stdio.h>
#endif

#include <errno.h>
#include <string.h>

#include "hirestime.h"
#include "lib_conversions.h"


SerialFTDI::SerialFTDI(const char*  usb_descr_string,
                       const char*  usb_serial_string,
                       int          baudrate,
                       std::string  specialopts) : Serial()
{
#if DEBUG > 1
  printf("SerialFTDI::SerialFTDI(\"%s\", \"%s\", %i)\n", usb_descr_string, usb_serial_string, baudrate);
#endif

  int nLatency       = -1;
  int nReadChunksize = -1;
  int nPID           = 0x6001;
  
  std::string::size_type pos1;  
  pos1 = specialopts.find(",ftdi_ReadChunksize=");
  if (pos1 != std::string::npos)
    nReadChunksize = atoi(specialopts.substr(pos1+20, specialopts.find(",", pos1+20)-pos1-20).c_str());
  pos1 = specialopts.find(",ftdi_Latency=");
  if (pos1 != std::string::npos)
    nLatency = atoi(specialopts.substr(pos1+14, specialopts.find(",", pos1+14)-pos1-14).c_str());
  pos1 = specialopts.find(",ftdi_PID=0x");
  if (pos1 != std::string::npos)
    nPID = hex2int(specialopts.substr(pos1+12, specialopts.find(",", pos1+12)-pos1-12));
  
  
#if DEBUG > 1
  printf("nReadChunksize=%i\n", nReadChunksize);
  printf("nLatency=%i\n",       nLatency);
  printf("Product ID=%i\n",     nPID);
#endif
    
  int nCode;
  
  if ((nCode = ftdi_init(&ftdi)) != 0)
    SetErr("ftdi_init: " + itoStr(nCode, ' ', 1) + " " + ftdi_get_error_string(&ftdi));
  
  // Der folgende Aufruf ist überflüssig, weil libftdi das selbst macht, wenn man das 
  // USB-Device öffnet.
  // if (IsOK())
  //   if ((nCode = ftdi_set_interface(&ftdi, INTERFACE_ANY)) != 0)
  //     SetErr("ftdi_set_interface: " + itoStr(nCode, ' ', 1) + " " + ftdi_get_error_string(&ftdi));
  
  if (IsOK())
    if ((nCode = ftdi_usb_open_desc(&ftdi, 0x0403, nPID, usb_descr_string, usb_serial_string)) != 0)
      SetErr("ftdi_usb_open_desc: " + itoStr(nCode, ' ', 1) + " " + ftdi_get_error_string(&ftdi));

  // ftdi_usb_reset wird bereits durch den obigen Aufruf gemacht
    
  // Wie ist der Wert normiert? Das Datenblatt sagt, man könne zwischen 1ms und 255ms einstellen; 
  // ftdi.c (libftdi) akzeptiert einen Wertebereich von 1 bis 255. Also ist 1LSB<=>1Millisekunde.
  // Der aktuelle Kernel (2.6.18) stellt von sich aus ein...16ms, würde ich sagen. Im Treiber
  // gibt es die Zeile
  //     static DEVICE_ATTR(latency_timer, S_IWUSR | S_IRUGO, show_latency_timer, store_latency_timer);
  // und eine Funktion static void create_sysfs_attrs(struct usb_serial *serial).
  // Also sucht man unter /sys/ nach einer Datei mit dem Namen latency_timer. Sie hat den Inhalt 16.
  // Also wird wohl 16ms eingestellt. Ist auch verständlich, weil die alten FTDI-Bausteine (A) fest auf 
  // 16ms eingestellt waren. Wenn man 16ms oder mehr einstellt, ist das also sicher nicht verkehrt.
  // Was steht drin, wenn man nichts einstellt?
  {
    unsigned char lmax = 16;
    unsigned char latency;
    
    if (nLatency < lmax)
      lmax = nLatency;
    
    if (IsOK())
      if ((nCode = ftdi_get_latency_timer(&ftdi, &latency)) != 0)
        SetErr("ftdi_get_latency_timer: " + itoStr(nCode, ' ', 1) + " " + ftdi_get_error_string(&ftdi));

#if DEBUG > 1
    printf("Per default steht die Latenzzeit auf %i ms\n", latency);
#endif
        
    if (latency > lmax && IsOK())
      if ((nCode = ftdi_set_latency_timer(&ftdi, lmax)) != 0)
        SetErr("ftdi_set_latency_timer: " + itoStr(nCode, ' ', 1) + " " + ftdi_get_error_string(&ftdi));
  }

  if (IsOK() && nReadChunksize > 0)
    if ((nCode = ftdi_read_data_set_chunksize(&ftdi, nReadChunksize)) != 0)
      SetErr("ftdi_read_data_set_chunksize: " + itoStr(nCode, ' ', 1) + " " + ftdi_get_error_string(&ftdi));
  
  if (IsOK())
    if ((nCode = ftdi_set_line_property(&ftdi, BITS_8, STOP_BIT_1, NONE)) != 0)
      SetErr("ftdi_set_line_property: " + itoStr(nCode, ' ', 1) + " " + ftdi_get_error_string(&ftdi));
  
  ChangeBaudrate(baudrate);
  
  Flush();
}

SerialFTDI::~SerialFTDI()
{
#if DEBUG > 1
  printf("SerialFTDI::~SerialFTDI(): %s\n", GetFirstErr().c_str());
#endif
  
  int nCode;
    
  if (IsOK())
    if ((nCode = ftdi_usb_close(&ftdi)) != 0)
      SetErr("ftdi_close: " + itoStr(nCode, ' ', 1) + " " + ftdi_get_error_string(&ftdi));
  
  ftdi_deinit(&ftdi);
}

void SerialFTDI::WriteByte(unsigned char uchVal)
{
  int nCode;
    
  if (IsOK())
    if ((nCode = ftdi_write_data(&ftdi, &uchVal, 1)) < 1)
      SetErr("ftdi_write_data: " + itoStr(nCode, ' ', 1) + " " + ftdi_get_error_string(&ftdi));
}

int SerialFTDI::WriteBytes(const unsigned char* buf,
                           int                  nBufLen)
{
  // Sorrily the data parameter for ftdi_write_data and usb_bulk_write is not const, 
  // which forces me to copy here? I don't want to give up this "const"...
  unsigned char* lbuf = (unsigned char*)malloc(nBufLen);
  int nCode = -1;
  
  if (lbuf != 0)
  {
    memcpy(lbuf, buf, nBufLen);
      
    if (IsOK())
      if ((nCode = ftdi_write_data(&ftdi, lbuf, nBufLen)) < nBufLen)
        SetErr("ftdi_write_data: " + itoStr(nCode, ' ', 1) + " " + ftdi_get_error_string(&ftdi));
    
    free(lbuf);
  }
  
  return(nCode);
}

int SerialFTDI::ReadByte(bool fSetErr,
                         int  nTimeout)
{
  unsigned char data;
  int nCode;
  HiResTime hrt;
  hrt.Capture();
  
  while (hrt.GetUSecSinceCapture() < nTimeout)
  {
    nCode = ReadByte(data);
    switch (nCode)
    {
      case 1:
        return(data);
      case -1:
        if (fSetErr)
          SetErr("error reading from port (could not read byte)");
        return(-1);
      default:
        break;
    }
    HiResTime::sleep_us(1000);
  }
  
  if (fSetErr)
    SetErr("error reading from port (could not read byte; timeout!)");
  
  return(-1); // timeout
}

int SerialFTDI::ReadByte(unsigned char& data)
{
  int nCode = -1;
        
  if (IsOK())
    if ((nCode = ftdi_read_data(&ftdi, &data, 1)) < 0)
      SetErr("ftdi_read_data: " + itoStr(nCode, ' ', 1) + " " + ftdi_get_error_string(&ftdi));
      
  return(nCode);
}

int SerialFTDI::ReadBytes(unsigned char* databuf, int nMaxBytesToRead)
{
  int nCode = -1;

  if (IsOK())
    if ((nCode = ftdi_read_data(&ftdi, databuf, nMaxBytesToRead)) < 0)
      SetErr("ftdi_read_data: " + itoStr(nCode, ' ', 1) + " " + ftdi_get_error_string(&ftdi));
      
  return(nCode);
}

void SerialFTDI::Flush()
{
  int nCode;
  
  if (IsOK())
    if ((nCode = ftdi_usb_purge_buffers(&ftdi)) != 0)
      SetErr("ftdi_usb_purge_buffers: " + itoStr(nCode, ' ', 1) + " " + ftdi_get_error_string(&ftdi));
}

void SerialFTDI::WaitUntilWritten()
{
  // ?
}

void SerialFTDI::ChangeBaudrate(int baudrate)
{
  int nCode;
  
  if (IsOK())
  {
    if ((nCode = ftdi_set_baudrate(&ftdi, baudrate)) != 0)    
      SetErr("ftdi_set_baudrate: " + itoStr(nCode, ' ', 1) + " " + ftdi_get_error_string(&ftdi));
    else
      nBaudrate = baudrate;
  }
}

void SerialFTDI::SetReadChunkSize(unsigned int chunksize)
{
  ftdi_read_data_set_chunksize(&ftdi, chunksize);
}

