/*
 * Copyright (C) 2007  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "serial_posix.h"

// 0: keine Ausgaben
// 1: nur Fehler
// 2: weiteres
#define DEBUG 0

#if DEBUG > 0
# include <stdio.h>
#endif

#include <errno.h>
#include <string.h>

#include <fcntl.h>   // seriell
#include <unistd.h>  // isatty

#include "hirestime.h"

SerialPOSIX::SerialPOSIX(const char*  sd,
                         int          baudrate) : Serial()
{
#ifdef WIN32
  SetErr("nicht implementiert für WIN32\n");
#else
  
#if DEBUG > 1
  HiResTime hrt;
  hrt.Capture();
#endif
  
  serPort = open(sd, O_RDWR | O_NOCTTY | O_NONBLOCK );

#if DEBUG > 1
  struct timeval diff;
  hrt.GetTimeSinceCapture(&diff);  
  printf("Open: %i s, %i µs\n", (int)(diff.tv_sec), (int)(diff.tv_usec));
  // todo: sometimes this takes 22s!
#endif
  
  if (serPort < 0)
  {
    SetErr("Fehler beim Öffnen des Ports\n");
  }
  
  Set_input_mode (baudrate);  
#endif
  Flush();
}

SerialPOSIX::~SerialPOSIX()
{
  printf("SerialPOSIX::~SerialPOSIX()\n");
  Reset_input_mode();
#ifndef WIN32  
  if (close(serPort))
  {
    printf("Fehler beim Schliessen des Ports\n");
    // todo: das bekommt niemand mehr mit...
    SetErr("Fehler beim Schliessen des Ports\n");
  }
#endif
#if DEBUG > 1
  printf("SerialPOSIX::~SerialPOSIX(): %s\n", GetFirstErr().c_str());
#endif    
}

void SerialPOSIX::WriteByte(unsigned char uchVal)
{
#ifndef WIN32
  int retcode;
  
  do 
  {
    retcode = write(serPort, &uchVal, 1);
    if (retcode == -1 && errno == EAGAIN)
    {
      HiResTime::sleep_us(1000);
      retcode = 0;
    }
  }
  while (retcode == 0);
#endif
}

int SerialPOSIX::WriteBytes(const unsigned char* buf,
                            int                  nBufLen)
{
#ifndef WIN32
  int retcode;
  int nWritten = 0;
  
  do 
  {
    retcode = write(serPort, buf+nWritten, nBufLen-nWritten);
    
    if (retcode > 0)
      nWritten += retcode;
    
    if (retcode == -1)
    {
      if (errno == EAGAIN)
        HiResTime::sleep_us(1000);
      else
      {
        SetErr(strerror(errno));
        break;
      }
    }
  }
  while (nWritten < nBufLen);
  
  return(nWritten);
#else
  return(0);
#endif
}

int SerialPOSIX::ReadByte(bool fSetErr,
                          int  nTimeout)
{
#if 0
  // Problem with this code: select might be interrupted by a signal and "man 2 select"
  // tells not to rely on the changed timeout value. So you can't have a well defined timeout
  // when there's a chance of being interrupted by a signal.
  unsigned char uchVal;
  int           cnt2;
  
  fd_set         set;
  struct timeval timeout;

  timeout.tv_sec  = 0;
  timeout.tv_usec = nTimeout;

  /* Initialize the file descriptor set. */
  FD_ZERO (&set);
  FD_SET (serPort, &set);

  //
  select(FD_SETSIZE, &set, NULL, NULL, &timeout);
  cnt2 = ::read(serPort, &uchVal, 1);

  if (cnt2 != 1)
  {
    std::string msg = "";

    if (cnt2 == -1)
      msg = (std::string)(": ") + (std::string)(strerror(errno));
      
    if (fSetErr)
      SetErr("error reading from port (could not read byte, timeout?)" + msg);
    return(-1);
  }
  else
    return((int)uchVal);
#else
  unsigned char data;
  int nCode;
  HiResTime hrt;
  hrt.Capture();
  
  while (hrt.GetUSecSinceCapture() < nTimeout)
  {
    nCode = ReadByte(data);
    switch (nCode)
    {
      case 1:
        return(data);
      case -1:
        if (fSetErr)
          SetErr("error reading from port (could not read byte)");
        return(-1);
      default:
        break;
    }
    HiResTime::sleep_us(1000);
  }
  
  if (fSetErr)
    SetErr("error reading from port (could not read byte; timeout!)");
  
  return(-1); // timeout  
#endif
}

int SerialPOSIX::ReadByte(unsigned char& data)
{
#ifdef WIN32
  return(0);
#else
  int nRetcode = ::read(serPort, &data, 1);
  
  if (nRetcode == -1)
  {
    if (errno == EAGAIN || errno == EINTR)
    {
      // This is not an error. I do only return a byte if there is one.
      // So maybe next time there will be a byte...
      nRetcode = 0;
    }
    else
    {
      SetErr(strerror(errno));
    }
  }
  
  return(nRetcode);
#endif
}

void SerialPOSIX::Flush()
{
#if DEBUG > 1
  printf("void SerialPOSIX::Flush()\n");
#endif
    
#ifndef WIN32
      
# if 0
    
  if (tcflush(serPort, TCIOFLUSH) != 0)
  {
    SetErr((std::string)"tcflush: " + strerror(errno));
#  if DEBUG > 0
    perror("tcflush");
#  endif    
  }
    
# else
  const int buflen = 256;
  char buf[buflen];
  int code;
  do
  {
    code = ::read(serPort, buf, buflen);
    if (code < buflen && (errno == EINTR || errno == EINTR))
      code = buflen;
  }
  while (code == buflen);
  
# endif
    
#endif
}

void SerialPOSIX::WaitUntilWritten()
{
#ifndef WIN32
  if (tcdrain(serPort) != 0)
  {
    SetErr((std::string)"tcdrain: " + strerror(errno));
# if DEBUG > 0
    perror("tcdrain");
# endif    
  }
#endif  
}

void SerialPOSIX::Reset_input_mode()
{
#ifndef WIN32
  if (tcsetattr (serPort, TCSANOW, &saved_attributes) != 0)
  {
    SetErr((std::string)"tcsetattr: " + strerror(errno));
# if DEBUG > 0
    perror("tcsetattr");
# endif    
  }
#endif
}

void SerialPOSIX::Set_input_mode(int baudrate)
{
#if DEBUG > 1
  printf("void SerialPOSIX::Set_input_mode(%i)\n", baudrate);
#endif
#ifndef WIN32
  // Make sure serPort is a terminal.
  if (!isatty (serPort))
  {
    SetErr("Not a terminal.\n");
  }

  if (IsOK())
  {
    // Save the terminal attributes so we can restore them later.
    if (tcgetattr (serPort, &saved_attributes) != 0)
    {
      SetErr((std::string)"tcsetattr: " + strerror(errno));
# if DEBUG > 0
      perror("tcsetattr");
# endif 
    }

    // Set the funny terminal modes. Siehe "man termios" !!
    if (tcgetattr (serPort, &tattr) != 0)
    {
      SetErr((std::string)"tcsetattr: " + strerror(errno));
# if DEBUG > 0
      perror("tcsetattr");
# endif 
    }
            
    tattr.c_iflag = 0;
    tattr.c_oflag = 0;
    tattr.c_cflag = CS8 | CREAD | CLOCAL;
    tattr.c_lflag = 0;
    tattr.c_cc[VMIN]  = 0;
    tattr.c_cc[VTIME] = 0;    
        
    ChangeBaudrate(baudrate);
  }
#endif
}

void SerialPOSIX::ChangeBaudrate(int baudrate)
{
#if DEBUG > 1
  printf("void SerialPOSIX::ChangeBaudrate(%i)\n", baudrate);
#endif
#ifndef WIN32
  speed_t speed;
  
  nBaudrate = baudrate;
  switch (baudrate)
  {
    case 0:
      speed = B50;
      nBaudrate = 50;
      break;
    case 75:
      speed = B75;
      break;
    case 110:
      speed = B110;
      break;
    case 134:
      speed = B134;
      break;
    case 150:
      speed = B150;
      break;
    case 200:
      speed = B200;
      break;
    case 300:
      speed = B300;
      break;
    case 600:
      speed = B600;
      break;
    case 1200:
      speed = B1200;
      break;
    case 1800:
      speed = B1800;
      break;
    case 2400:
      speed = B2400;
      break;
    case 4800:
      speed = B4800;
      break;
    case 9600:
      speed = B9600;
      break;
    case 19200:
      speed = B19200;
      break;
    case 38400:
      speed = B38400;
      break;
    case 57600:
      speed = B57600;
      break;
    case 115200:
      speed = B115200;
      break;
    case 230400:
      speed = B230400;
      break;
    default:
      SetErr("Illegal baudrate\n");
      speed     = B19200;
      nBaudrate = 19200;
      break;
  }
  
  cfsetospeed(&tattr, speed);   
  cfsetispeed(&tattr, speed);   
  
  if (tcsetattr (serPort, TCSANOW, &tattr) != 0)
  {
    SetErr((std::string)"tcsetattr: " + strerror(errno));
# if DEBUG > 0
    perror("tcsetattr");
# endif
  }
#endif
}

