// -*- mode: c; mode: fold -*-

/*
 * Copyright (C) 2005-2008  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "config.h"

#include "usrcmd.h"

#include "lib_conversions.h"
#include <iostream>
#include <cstdlib>

UserVars* UserCmd::vars;

void UserCmd::WriteValueToVarCache(int val)
{
  if (myVarname.length())
    vars->Write(myVarname, val);
}

std::string UserCmd::GetVarName(std::string in, std::string::size_type& pos)
{ 
  if (in[pos] != '$')
  {
    std::cerr << "Expected '$' at the beginning of " << in.substr(pos) << "\n";
    throw std::exception();
  }
   
  std::string tmp;
  std::string::size_type end = pos+1;
  while (end < in.length() && 
         (isalnum(in[end]) || in[end] == '_')
         )
    end++;
  
  tmp = in.substr(pos, end-pos);
  pos = end;
  return(tmp);  
}

int UserCmd::setCmd(std::string cmdstr)
{
  std::string origline = cmdstr;
  
  fErr  = false;
  error = "";
  
  myVarname = "";

  // remove whitespace
  {
    std::string tmp = "";
    for (unsigned int n=0; n<cmdstr.length() && cmdstr[n] != ';' && cmdstr[n] != '#'; n++)
    {
      if (isspace(cmdstr[n]) == 0)
        tmp += cmdstr[n];
    }
    cmdstr = tmp;
  }
  
  if (cmdstr.length() < 1)                                 
  {
    explanation = "ignore this line";
    myCmd = eNOP;
    /*
    fErr  = true;
    error = "Command too short";
    return(-1);
     */
  }
  
  else if (cmdstr.compare("loop") == 0)                    
  {
    explanation = "start from beginning of macro file";
    myCmd       = eLOOP;
  }
  
  else if (cmdstr.compare("eof") == 0)                     
  {
    explanation = "stop reading from macro file; enter interactive mode";
    myCmd = eEOF;
  }
  
  else if (cmdstr.compare("newline") == 0)                 
  {
    explanation = "new output line in table mode";
    myCmd = eNEWLINE;
  }
  
  else if (cmdstr.compare("q") == 0)                       
  {
    explanation = "quit the program";
    myCmd = eQUIT;
  }
  
  else if (cmdstr[0] == 's' && cmdstr.length() > 1)        
  {
    nVal = atoi(cmdstr.substr(1).c_str());
    explanation = ((std::string)"sleep for ") + itoStr(nVal, ' ', 1) + ((std::string)" µs");
    myCmd = eSLEEP;
  }
  
  else if (cmdstr.find("e\"") == 0)                        
  {
    explanation = "echo string";
    myCmd = eECHO;
    std::string::size_type pos1 = origline.find('"')+1;
    std::string::size_type pos2 = origline.rfind('"');
    echo  = origline.substr(pos1, pos2-pos1);
  }
  
  else if (cmdstr[0] == '#' || cmdstr[0] == ';')           
  {
    explanation = "ignore this line";
    myCmd = eNOP;
  }
  
  else if (cmdstr[0] == 'h')                               
  {
    explanation = "print help";
    myCmd = eHELP;
  }
  
  else if (cmdstr[0] == 'f' && cmdstr.length() > 1)        
  {
    myCmd    = eREADFILE;
    filename = cmdstr.substr(1);
    explanation = "read file " + filename;
  }
  
  else // read, write, until
  {
    nWidth = 0; // to be determined

    if (cmdstr[0] == '$')
    {
      std::string::size_type pos     = 0;
      std::string            varname = GetVarName(cmdstr, pos);
      if (cmdstr[pos] != '=')
      {
        setErr("Expected '=' after " + varname + " at the beginning of " + cmdstr);
      }
      else
      {
        // std::cerr << varname << " wird beschrieben\n";
        myVarname = varname;
        cmdstr = cmdstr.substr(pos+1);
      }
    }
    
    
    // --- Extract numerical arguments ------------------------------
    // There can be up to four numbers in the string, separated by
    // non-number digits.
    const unsigned int nNrMax = 4;
    int           nNr [nNrMax]; // number
    char          cSep[nNrMax]; // non-number character in front of number
    int           nFCA[nNrMax]; // index of first character after number
    unsigned int  nNrUsed = 0;  // arguments consumed so far
    std::string   optstr;
    unsigned int  nNrCnt  = 0;
    {
      bool fNr   = false;
      int  start = 0; // first index of current number

      for (unsigned int n=0; n<cmdstr.length(); )
      {
        if (cmdstr[n] == '$')
        {
          std::string::size_type pos     = n;
          std::string            varname = GetVarName(cmdstr, pos);
          
          if (nNrCnt == nNrMax)
          {
            setErr("too many numbers found in " + cmdstr);
          }
          else
          {
            if (n>0)
              cSep[nNrCnt] = cmdstr[n-1];
            else
              cSep[nNrCnt] = ' ';

            nFCA[nNrCnt]   = pos;
            // todo: reading variables here may cause strange effects: reading old value when
            // parsing block of commands etc. Before using a variable, the pipeline should be
            // flushed...
            nNr [nNrCnt++] = vars->Read(varname);
          }
          
          n = pos;
        }
        else
        {
          if (cmdstr[n] < '0' || cmdstr[n] > '9')
          {
            if (fNr)
            {
              if (nNrCnt == nNrMax)
              {
                setErr("too many numbers found in " + cmdstr);
              }
              else
              {
                nFCA[nNrCnt] = n;
                nNr [nNrCnt++] = atoi(cmdstr.substr(start, n-1).c_str());
                fNr = false;
              }
            }
          }
          else if (fNr == false)
          {
            start = n;
            fNr = true;
            if (n>0)
              cSep[nNrCnt] = cmdstr[n-1];
            else
              cSep[nNrCnt] = ' ';
          }
          
          n++;
        }
      }

      if (fNr && nNrCnt < nNrMax)
      {
        nFCA[nNrCnt]   = -1;
        nNr [nNrCnt++] = atoi(cmdstr.substr(start).c_str());
      }
    }
    
#if 1==0
    for (unsigned int n=0; n<nNrCnt; n++)
    {
      std::cerr << " '" << cSep[n] << "' " << nNr[n];
    }
    std::cerr << "\n";
#endif
    
    // Is 'number of bits per value' given?
    if (nNrCnt > 1 &&
        cSep[1] != '-' && cSep[1] != '*' &&                     // range
        cSep[1] != ':' &&                                       // write
        cSep[1] != '=' && cSep[1] != '>' && cSep[1] != '<'      // wait until
        )
    {
      // So there is an 'optional' (in this case mandatory) argument after the first number,
      // which seems to be 'number of bits per value'.
      // std::cerr << nNrUsed << " " << nFCA[nNrUsed] << "\n";
      optstr = cmdstr.substr(nFCA[nNrUsed]);
      nWidth = nNr [nNrUsed++];      
    }
    else
    {
      optstr = cmdstr.substr(1);
    }

    // 'number of bits per value' via 'w'?
    //std::cerr << "Options: " << optstr << "\n";
    if (optstr.find('w') != std::string::npos)
    {
      if (nWidth == 0)
        nWidth = 16;
      else
        setErr("NumberOfBits given by number and w: " + cmdstr);
    }
    else
    {
      if (nWidth == 0)
        nWidth = 8;
    }
    //std::cerr << "Number of bits per value: " << nWidth << "\n";
    
    // Next is address (range)
    nSize = nWidth/8;
    nAdr = nNr[nNrUsed++];
    if (nNrUsed < nNrCnt)
    {
      if (cSep[nNrUsed] == '-')
      {
        if (nNr[nNrUsed] - nAdr < 0)
        {
          setErr("Address range reversed. There may be an error: " + cmdstr);
        }
        nSize = nNr[nNrUsed] - nAdr + (nWidth/8);
        if (nSize % (nWidth/8) != 0)
          setErr("illegal size of range in " + cmdstr);
        nNrUsed++;
      }
      else if (cSep[nNrUsed] == '*')
      {
        nSize = nNr[nNrUsed++] * (nWidth/8);
      }
    }
    //std::cerr << "Number of bytes to be transferred: " << nSize << "\n";
    
    // --- Extract command ------------------------------------------
    switch (cmdstr[0])
    {
      case 'u':
        {
          myCmd  = eWAIT;
          
          if (nSize*8 != nWidth)
            setErr("Only one value can be evaluated: " + cmdstr);

          if (nNrCnt - nNrUsed != 1)
            setErr("Wrong number of arguments: " + cmdstr);
            
          nVal = nNr[nNrUsed++];
          
          explanation = "wait until the value at address " + itoStr(nAdr, ' ', 1);
          
          if (cmdstr.find("==") != std::string::npos)
          {
            explanation += " is equal to ";
            myRelation = eEQUAL;
          }
          else if (cmdstr.find("!=") != std::string::npos)
          {
            explanation += " is different from ";
            myRelation = eNOTEQUAL;
          }
          else if (cmdstr.find(">") != std::string::npos)
          {
            explanation += " is greater than ";
            myRelation = eGREATER;
          }
          else if (cmdstr.find("<") != std::string::npos)
          {
            explanation += " is smaller than ";
            myRelation = eSMALLER;
          }
          else
          {
            setErr("unknown relation in " + cmdstr);
          }          
          explanation += itoStr(nVal, ' ', 1);
        }
        break;
        
      case 'r':
        {
          myCmd  = eRW;
          fWrite = false;
          explanation += " read from addr " + itoStr(nAdr, ' ', 1) + " ";          

          if (nNrCnt - nNrUsed != 0)
            setErr("Wrong number of arguments: " + cmdstr);
        }
        break;
        
      case 'w':
        {
          myCmd  = eRW;
          fWrite = true;
          
          if (nNrCnt - nNrUsed != 1)
            setErr("Wrong number of arguments: " + cmdstr);
          
          nVal = nNr[nNrUsed++];
          explanation += " write from addr " + itoStr(nAdr, ' ', 1) + " ";          
          explanation += itoStr(nSize/(nWidth/8), ' ', 1) + " value(s) to " + itoStr(nVal, ' ', 1);                    
        }
        break;
        
      default:
        {
          setErr("unknown command: " + cmdstr);
        }        
        break;
    }    
        
    // --- Extract how to display result ---------------------------- 
    if (optstr.find("s") != std::string::npos)
    {
      explanation = "(signed   output) ";
      myDisplay = eSIGNED;
    }
    else if (optstr.find("u") != std::string::npos)
    {
      explanation = "(unsigned output) ";
      myDisplay = eUNSIGNED;
    }
    else if (optstr.find("h") != std::string::npos)
    {
      explanation = "(hex      output) ";
      myDisplay = eHEX;
    }
    else
    {
      explanation = "(full     output) ";
      myDisplay = eALL;
    }
    
    // --- Extract whether result should be shown ------------------- 
    if (optstr.find("q") != std::string::npos)
    {
      myShowValue = eQUIET;
    }
    else if (optstr.find("Q") != std::string::npos)
    {
      myShowValue = eREALLY_QUIET;
    }
    else
      myShowValue = eNORMAL;    
  }
  
  
#if (GLOB_DEBUG > 0)
  std::cout << "User command: " << explanation << "\n";
#endif

  if (fErr)
    return(-1);
  else
    return(0);
}

void UserCmd::setErr(std::string descr)
{
  if (!fErr)
    error = descr;
  fErr = true;
}

